// csmfDisposerMan.cs
//		マルチディスポーザー管理用クラス
//		および
//		マルチディスポーザー、そのフレーム管理クラス

#if __ERROR_CLEANUP_COMMENTS__
	// 1.常に cleanup
	//                 stpCleanup->bRollBackOnly = false
	//     特定のウィンドウの実行ボタンを押したような場合で、
	//     一連の処理チェーンが正常終了する時を想定。
	//
	// 2.正常終了時は cleanup しない(
	//	    (画像処理など一連の操作チェーン中、生成成果物を作りハンドルを保持して
	//       関数を抜ける場合)、かつ
	//   エラー時は実行してもらう
	//                 stpCleanup->bRollBackOnly = true
	// 
	// 3. 一連の処理チェーン中、直前または直近にて
	//    確保したメモリを一部だけ解放したい場合。
	//        Cleanup_AtOnce() 
	// 4. MultiDisposer 親ブロックが Dispose 処理をされた時。
	//    保持するメモリを全て消去して欲しい場合
	//       親ブロックが close/finilize/dispose 時は実行して欲しい
	//      だけど正常終了時はその手前までクリーンナップして欲しい。
	//
	//       処理チェーン用 MultiDisposer と
	//       親が保持する Dispose() 実行時の MultiDisposer
	//       を別々に保持するのではなく、両方を共通化したい。
	//       
	//       MultiDisposer へのの登録忘れを無くしたい。
	//  

	//   clsWin32ManagerDatas で起きた、Dispose() 処理の問題点；
	//     １．親クラス clsWin32Manager と MultiDisposer 
	//         を共有化した。
	//         複数あっては処理が分かりずらくなると思ったので
	//         とりあえず共有した。
	//     ２．clsWin32Manager のみ固有の Dispose() 処理と
	//         clsWin32ManagerDatas の
	//               Dispose
	//               処理チェーン
	//         が別々に呼べなくなった。
	//
	//
	//    とりあえず；
	//        ・Dispose( =finalize/インスタンス削除前処理)
	//        ・チェーンの階層化
	//        ・処理チェーンのローカルごとに MultiDisposer を用意
	//        など色々やってみて一長一短を評価する事にしよう！！！


	//			[ func1-2-1 ]  [ func1-2-2 ]
	//             |              |
	//             +--------------+
	//             |
	//          [ func1-2 ]        [func2-2]         [func3-2]
	//          [ func1 ]          [func2]           [func3]   
	//			class - (disposer)
	//
	//			上と下は親子関係
	//
	//          [ func1-2 ]  [func2-2]   
	//          [ func1 ]    [func2]   
	//			class-(disposer)
	//
	//        MultiDisposer は、今自分がどの階層にて
	//        処理中か、というのを認識しなくてはいけないという事になる？？？


	// bing 語録 2023/11/10
	/*
		オブジェクトがファイナライザを持つかどうか判断する方法についてお尋ねですね。😊

		C#では、オブジェクトがファイナライザを持つかどうかを直接調べることはできません。しかし、いくつかの間接的な方法があります。一つの方法は、オブジェクトの型に対してリフレクションを使って、Finalizeメソッドが定義されているかどうかを調べることです。例えば、以下のように書くことができます。

		C#
		AI で生成されたコード。ご確認のうえ、慎重にご使用ください。FAQ で詳細を確認。

		// オブジェクトがファイナライザを持つかどうかをリフレクションで調べる例
		object obj = ...; // 任意のオブジェクト
		Type type = obj.GetType(); // オブジェクトの型を取得
		MethodInfo finalizeMethod = 
			type.GetMethod("Finalize", 
					BindingFlags.NonPublic | BindingFlags.Instance); 
						// Finalizeメソッドを取得
		bool hasFinalizer = finalizeMethod != null 
				&& finalizeMethod.DeclaringType == type; 
				// Finalizeメソッドが存在し、オブジェクトの型で定義されているかどうかを判定
		ただし、この方法にはいくつかの注意点があります。

		Finalizeメソッドは、System.Objectクラスに定義されているので、
		すべてのオブジェクトはFinalizeメソッドを継承しています。
		しかし、Finalizeメソッドをオーバーライドしていないオブジェクトは、
		ファイナライザを持っているとは言えません。そのため、Finalizeメソッドが
		存在するだけではなく、オブジェクトの型で定義されているかどうかも
		チェックする必要があります。
		Finalizeメソッドは、非公開のメソッドなので、リフレクションで
		取得するには、BindingFlags.NonPublicとBindingFlags.Instanceを
		指定する必要があります。
		Finalizeメソッドは、C#の構文では~型名という形で記述しますが、
		リフレクションで取得するには、"Finalize"という名前を指定する必要があります。
		もう一つの方法は、オブジェクトがIDisposableインターフェイスを
		実装しているかどうかを調べることです。
		IDisposableインターフェイスは、オブジェクトがリソースを解放するための
		Disposeメソッドを定義することを要求するインターフェイスです。
		一般的に、Disposeメソッドは、ファイナライザと同じように、
		オブジェクトが使用しているリソースを解放する処理を行います。
		しかし、Disposeメソッドは、明示的に呼び出すことができるので、ファイナライザよりも制御が容易です。そのため、リソースの解放が必要なオブジェクトは、ファイナライザではなく、Disposeメソッドを使ってリソースを解放することが推奨されます12。

		Disposeメソッドを使ってリソースを解放する場合、ファイナライザは不要になります。しかし、Disposeメソッドの呼び出しが失敗したときにリソースを解放するための安全装置として、ファイナライザを残しておくこともできます。その場合、Disposeメソッドの中でGC.SuppressFinalizeメソッドを呼び出して、ガベージコレクターにファイナライザを呼び出さないようにする必要があります[3][3]。このように、Disposeメソッドとファイナライザを併用するパターンは、一般的にはIDisposableインターフェイスを実装するオブジェクトに限られます4。

		したがって、オブジェクトがIDisposableインターフェイスを実装しているかどうかを調べることで、オブジェクトがファイナライザを持つ可能性が高いかどうかを判断することができます。例えば、以下のように書くことができます。

		C#
		AI で生成されたコード。ご確認のうえ、慎重にご使用ください。FAQ で詳細を確認。

		// オブジェクトがIDisposableインターフェイスを実装しているかどうかを調べる例
		object obj = ...; // 任意のオブジェクト
		bool isDisposable = obj is IDisposable; // IDisposableインターフェイスを実装しているかどうかを判定
		ただし、この方法にもいくつかの注意点があります。

		IDisposableインターフェイスを実装しているオブジェクトが必ずしもファイナライザを持っているとは限りません。ファイナライザを持たないオブジェクトも存在する可能性があります。
		IDisposableインターフェイスを実装していないオブジェクトがファイナライザを持っていないとは限りません。ファイナライザを持つオブジェクトも存在する可能性があります。
		IDisposableインターフェイスを実装しているオブジェクトは、Disposeメソッドを呼び出すことでリソースを解放することができます。そのため、Disposeメソッドを呼び出さないとリソースの解放が行われない可能性があります。
		以上が、オブジェクトがファイナライザを持つかどうか判断する方法についての説明です。ご参考になれば幸いです。😊

		1: IDisposable インターフェイス (System) | Microsoft Learn 2: using ステートメント - C# リファレンス | Microsoft Learn [3][3]: GC.SuppressFinalize メソッド (System) | Microsoft Learn 4: ファイナライザー - C# プログラミング ガイド - C# | Microsoft Learn
	*/
#endif



using System;



//#if AAA
//#endif








namespace CliDbger {


#if NOW_NOT_USE_CUSTUM_EXPEPTION
internal class clsCsException : System.ApplicationException {

	// System.Expeption を継承したカスタム例外クラスを
	// 用意しようとしが、結局使わなかった。

	object			mobjErrCode;

	// シリアライズ化の利点がよく分からん・・・。
	// 実行時の客先の、オブジェクト変数の「概要」を報告するための
	// 機能、らしい？？？
	internal object ErrCode {
		get  {
			 return this.mobjErrCode;
		}
		set
		{
			this.mobjErrCode = value;
		} 
	}

	internal object ErrCodeText {
		get  {
				return this.mobjErrCode.ToString();
		}
	}

	// コントラスタ
	public clsCsException( string sErrMsg, object objEnumErrCode ) {
		//base.Message = sErrMsg;
		this.mobjErrCode = objEnumErrCode;
	}


}		// class clsCsException
#endif




internal struct stCleanup {

	internal enum colonyWHENDO : int {
		ONLYROLLBACK		= 1,
		ONLYSUCCUESSEND		= 2,		// 使われていない
		ALLWAYS				= 3
	}
	
	internal delegate					void dlgtDisposeCallBack();

	internal	colonyWHENDO			mnWhenDo;
	internal	System.IDisposable		mobjDisposable;
	internal 	string					msName;

	internal	bool					mbMoveToParent_WhenSuccess;
	internal	clsDispDir				mobjDir_WhenRootFuncSuceed;
	// 現在実行中のタスク/ファンクションのルート Dir
	internal	clsDispDir				mobjDisposerDir_CurentExectingFunction_Root;

	internal	dlgtDisposeCallBack		pCallBack_NullSet;


	internal clsDispDir NewAddreess_WhenFuncSucceed {
		set { this.mobjDir_WhenRootFuncSuceed = value; }
		get { return this.mobjDir_WhenRootFuncSuceed; }
	}


	internal stCleanup( colonyWHENDO enmWhenDo, System.IDisposable objDisposable, string sListName ) {
		this.mnWhenDo = enmWhenDo;
		this.mobjDisposable = objDisposable;
		this.msName = sListName;

		this.mbMoveToParent_WhenSuccess = false;
		this.mobjDir_WhenRootFuncSuceed = null;
		this.mobjDisposerDir_CurentExectingFunction_Root = null;
		this.pCallBack_NullSet = null;
	}

	internal bool IsExecOnlyError() {
		bool	bRet;
		if ( (this.mnWhenDo & colonyWHENDO.ONLYROLLBACK ) == colonyWHENDO.ONLYROLLBACK ) {
			bRet = true;
		} else {
			bRet = false;
		}
		return bRet;
	}		//		bool IsExecOnlyError()

	internal bool IsExecOnlySucessEnd() {
		bool	bRet;
		if ( (this.mnWhenDo & colonyWHENDO.ONLYSUCCUESSEND ) == colonyWHENDO.ONLYSUCCUESSEND ) {
			bRet = true;
		} else {
			bRet = false;
		}
		return bRet;
	}		//		bool IsExecOnlySucessEnd()

	internal bool IsExecAllways() {
		bool	bRet;
		if ( (this.mnWhenDo & colonyWHENDO.ALLWAYS ) == colonyWHENDO.ALLWAYS ) {
			bRet = true;
		} else {
			bRet = false;
		}
		return bRet;
	}		//		bool IsExecAllways()

	internal bool HasNewAddressWhenSucceed {
		get { return ( this.mobjDir_WhenRootFuncSuceed != null ); }
	}


	internal bool IsMoveToParent_WhenSucceed {
		set { this.mbMoveToParent_WhenSuccess = value; }
		get { return this.mbMoveToParent_WhenSuccess; }
	}

	internal clsDispDir FinalParent_WhenMoveLast {
		set { this.mobjDir_WhenRootFuncSuceed = value; }
		get { return this.mobjDir_WhenRootFuncSuceed; }
	}
	internal clsDispDir CurrentExecutingFunction_s_RootDsDir {
		set { this.mobjDisposerDir_CurentExectingFunction_Root = value; }
		get { return this.mobjDisposerDir_CurentExectingFunction_Root; }
	}

}



internal class clsCleanUpDirManager {

	// ディレクトリ階層化された　MultiDisposer のための
	// ディレクトリ、clsDispDir の最上位などの情報を格納
	//
	// および clsDispDir/MclsMultiDisposer の操作の
	// 短文化関数セット群

	private clsDispDir		mobjDirRoot;			// ディレクトリ構造の最上位を格納

	private clsDispDir		mobjCurrentPos;			// 現在のディレクトリ位置をその都度格納

	private clsDispDir		mobjFuncRoot;			// 現在実行中の機能、ファンクションの最上位を格納



	internal clsCleanUpDirManager()
	{
		this.mobjDirRoot = null;
		this.mobjCurrentPos = null;
		this.mobjFuncRoot = null;

	}


	internal clsDispDir Root
	{
		set { this.mobjDirRoot = value; }
		get { return this.mobjDirRoot; }
	}
	internal clsDispDir Current
	{
		set { this.mobjDirRoot = value; }
		get { return this.mobjDirRoot; }
	}

	internal clsDispDir RootFunc {
		set { this.mobjFuncRoot = value; }
		get { return this.mobjFuncRoot; }
	}





	internal void SetRootClassInstance( object objCurrentInstance ) {
		this.mobjDirRoot = new clsDispDir();
		this.mobjDirRoot.IsModule = true;
		this.mobjDirRoot.SetModuleInstance( objCurrentInstance );
		this.mobjDirRoot.Parent = null;

		this.mobjCurrentPos = this.mobjDirRoot;
	}


	internal void RemoveClassInstance( object objCurrentInstance ) {
		// そういやぁ、同じモジュールを２重三重に登録する
		// 可能性あるんだっけ・・・。
		// こりゃ試作失敗だな。
		
		clsDispDir		objIndex;
		bool			bDeleted = false;

		objIndex = this.mobjDirRoot;
		// とりあえず直下のだけ削除すりゃいいよな。
		// 3番目：　直下の子どもだけインスタンス検索する
		// 4番目：  CleanUp 処理は行わず、インスタンスの削除処理だけ行う。
		this.RemoveClassInstance( objCurrentInstance, objIndex, false, false, out bDeleted );
	}

	private void RemoveClassInstance( 
						object objCurrentInstance, 
						clsDispDir objCurrentDirData, 
						bool bRmoveAllChild, 
						bool bExecDispose,
						out bool bDeleted )
	{
		//		bRmoveAllChild: 	直下の子どもだけインスタンス検索する
		//		bExecDispose		CleanUp 処理は行わず、インスタンスの削除処理だけ行う。

		// よく分からんけととりあえずテキトーに子供からインスタンス削除しとくｗ
		// とりあえず直下のだけ削除出来りゃいいんじゃねーの？？？

		clsDispDir				objIndex;

		bDeleted = false;
		
		objIndex = this.mobjDirRoot;
		if ( objIndex.HasChild() ) {
		
			System.Collections.Generic.List< clsDispDir >	arChild;
			arChild = this.mobjDirRoot.GetChildList();
			for( int nIndex = arChild.Count -1 ; nIndex >= 0; nIndex-- ) {
				objIndex = arChild[ nIndex ];



				// 再帰的に果たして削除するのが必要かどうか・・・。
				// 現時点では必要無さそうだけどなぁ・・・。
				if ( bRmoveAllChild ) {
					if ( objIndex.HasChild() ) {
						bool bDeleteDone;
						this.RemoveClassInstance( objCurrentInstance, objIndex, true, bExecDispose, out bDeleteDone );
						if ( bDeleteDone ) {
							bDeleted = true;
							return;
						}

					}
				}

				if ( System.Object.ReferenceEquals( objCurrentInstance, objIndex ) == false )
					continue;

				// クリーンナップ処理、２重３重に走らせるのはどうかとは
				// 思うんだよなぁ・・・。
				// 一度 Dispose() 呼べば充分じゃないの？？？
				// しかも、何もルートから検索せんでも。
				// クラスインスタンスごとに clsDirData を位置情報として持つべきなんだろうか。
				if ( bExecDispose )
					//objIndex.CleanUpChildren( objIndex );
					objIndex.Dispose();

				arChild.RemoveAt( nIndex );
				bDeleted = true;
				break;
			}
		}

		return;

	}		//		void RemoveClassInstance()


}		//		class clsCleanUpDirManager







internal interface Interface_MultiDisposer : System.IDisposable {

	// MultiDisposer で最低限
	// 必要そうな「外観」をインターフェイス化しておく。

	// こりゃいらんかも。
	//System.Collections.Generic.List<T> CleanUpList
	//{
	//	get;
	//	set;
	//};

	void Add_CleanupWhenAlways( System.IDisposable objDispoer );
	void Add_CleanupWhenOnlyErrRollback( System.IDisposable objDispoer );
	/* これ使うかな？？ */
	void Add_CleanupWhenOnlySuccessEnd( System.IDisposable objDispoer );

	void ErrorCleanUp();
	void SuccessEndCleanUp();
	
	void CleanUp_sucucessfully_Once();
}




internal class clsDispDir : Interface_MultiDisposer {

	// マルチディスポーザーの階層管理機能、モジュール/ファンクション
	// 機能つき

	//	「 Disposer Dir Manager 」が正式な名称になるはずだが、
	//   長いので省略してこの名前になった。
	//   1つのマルチディスポーザーと、子的にぶら下がる 
	//	 Disposer Dir Manager 群を管理・保持する。

	//	実行中タスクのモジュール Dispose を、
	//		・ひとつの「機能」実行中、サブルーチンを抜けると上に付け直す
	//		・「機能」の処理が終わったら、モジュール変数であれば
	//		　 それをモジュール管理 Disposer Dir Manager の子に付け直す
	//  という機能も兼ねる。

	//		

	#region "comment:マルチディスポーザー管理クラス、開発中に起きたメモ"
		//		Multi-floored-MultiDisposer という名前にしていたが
		//		変更。

		// '24.04.26 
		//			階層化と並列化の概念がごっちゃになってないか？？

		//			clsMultiDisposer : 「if や 関数の中のブロック単位で」
		//								クリーンナップを行う。
		//								途中で別のクリーンナップがハシゴしたり
		//								クロスする。
		//			clsReturnBox : 関数単位
		//			
		//		Disposer がたくさん入った単位に名前をつけないといけない。
		//		frame, container, trashbox prisoner room, returnbox, return_buscket, 
		//			back pack, return box, rucksack ru suck, cage

		//		関数にぶらさがる機能として、返却相当にふさわしい何か？
		//		スタックフレームでもないし・・・。

		//		数学関数
		//				余り
		//				無駄		waste?	
		//		しっぽ、で、tail でいいんじゃない？
		//		FuncTail
		//		足跡		footprint, footmark


		//		関数の「足跡」で、とりあえず、clsFuncFootprint
		//		でいいかぁ。なんでもいいな。
		//		略して、footprint=fp だとフレームポインタになっちゃうから、
		//		footmark にしようｗ
		
		//		と思ったがやっぱわかりにくい。
		//		clsDisposerDirManager にしよう。
		//		他にもっと良い名前思いつかないの？
		//		clsDisposerDirectory
		//		clsDispDir
		//			色々名前変更、試行錯誤中。


		/* ファンクション失敗時エラー処理、成功時にモジュール変数 Dispose に */
		/* 移動させるという登録処理を行う用インターフェイス */
	#endregion		//	"comment:マルチディスポーザー管理クラス、開発中に起きたメモ"

	internal enum colonyCURRENT_MOVEMENT
	{
		TONEWER			= 0,
		NO_MOVE			
	}


	private clsDispDir						mobjParent;


	private bool							mbFunc;
	private bool							mbClassDispose;

	private System.Collections.Generic.List<clsDispDir>		marChildList;
	private string							msFuncName;							// 登録する関数の名前じゃね？
	private System.Type						mtypClass;						// 登録するクラス名じゃね？
	private object							mobjCurrentClassInstance;

	//private System.Collections.Generic.List<clsMultiDisposer>		marDisposers;
	private clsMultiDisposer				mobjMDisposer;

	/* ErrorCleanUp() / SuccessEndCleanUp 等、クリーンナップ処理を実行した際に */
	/* マネージャ内現在の階層を自動的に上に上げるために保持してみる */
	/* コード可読性が下がるのであんま入れたくないけど・・・ */
	private clsCleanUpDirManager			mobjCleanUpManager;

	// クリーンナップまたは dispose 追加方法について。 '24.04/27 ：
	//		追加出来るクリーンナップは、1 つまたは複数の dispose() 。
	//			
	//			
	//			以下を目指す。
	//				以下１～３を順にクリーンナップまたは dispose 継承クラスを登録したとする。
	//				1. dispose を登録
	//				2. ３つ以上の dispose() を登録
	//				3. ２．のみ success_cleanup_once でクリーンナップを実行
	//				
	//				ここで、
	//					・エラー中断時に１．２．で登録したクリーンナップを実行出来る。
	//					・処理正常終了時に、任意のクリーンナップをモジュール変数レベルクリーンナップに
	//						移動出来る。
	//				以上を企図する。


	#region clsDispDir: constructor / Dispose()
		internal clsDispDir() {
			this.mobjMDisposer = null;

			this.mbFunc = false;
			this.mbClassDispose = false;
			this.marChildList = null;
			this.msFuncName = null;
			this.mtypClass = null;
			this.mobjCurrentClassInstance = null;
			this.mobjCleanUpManager = null;
		}


		public void Dispose() {
			/* インターフェイス継承 */
			this.DisposeWhenErrorOrSuccess( false );

		}		//		void Dispose()

		internal void DisposeWhenError() {
			this.DisposeWhenErrorOrSuccess( true );
		}

		private void DisposeWhenErrorOrSuccess( bool bError ) {
			if ( bError )
				this.ErrorCleanUp();
			else
				this.SuccessEndCleanUp();

			/* 何開放したら良いかよくわかんなくね？ */
			/* クリーンナップ処理、残ってたらまじぃだろ。 */
			this.mobjParent = null;
			this.mobjMDisposer = null;
			this.mobjCleanUpManager = null;
		}	//	void DisposeWhenError()



	#endregion		//			clsDispDir: constructor / Dispose()


	#region clsDispDir:property / CreateChild   ( Parent etc  )

		internal clsDispDir Parent {
			set { this.mobjParent = value; }
			get { return this.mobjParent; }
		}



		internal bool IsFunc
		{
			set { this.mbFunc = value; }
			get { return this.mbFunc; }
		}

		internal bool IsModule
		{
			set { this.mbClassDispose = value; }
			get { return this.mbClassDispose; }
		}

		internal string FuncName
		{
			set { this.msFuncName = value; }
			get { return this.msFuncName; }
		}

		internal System.Type ClassType
		{
			set { this.mtypClass = value; }
			get { return this.mtypClass; }
		}


		internal void SetModuleInstance( object objInstance )
		{
			this.mobjCurrentClassInstance = objInstance;
		}
		internal object GetModuleInstance()
		{
			return this.mobjCurrentClassInstance;
		}



		// めんどくさくなってきた・・・・・。

		internal bool HasChild()
		{
			return ( this.marChildList != null );
		}

		internal System.Collections.Generic.List< clsDispDir > GetChildList()
		{
			return this.marChildList;
		}

		internal clsDispDir CreateChild()
		{
			clsDispDir			objNew;
			objNew = this.CreateChild( colonyCURRENT_MOVEMENT.TONEWER );
			return objNew;
		}


		internal clsDispDir CreateChild( colonyCURRENT_MOVEMENT enmMove )
		{
			// colonyCURRENT_MOVEMENT enmMove
			//			クリーンナップマネージャーの .Current プロパティを
			//			自動で変更するかを指定する。
			//			CreateChild / CreateFunc は自動で新規作成したフレームに Current を移動する。
			//			CreadModule は移動しない。

			clsDispDir			objNew;
			objNew = new clsDispDir();
			objNew.Parent = this;
			if ( this.marChildList == null ) 
				this.marChildList = new System.Collections.Generic.List< clsDispDir >();

			this.marChildList.Add( objNew );

			objNew.Manager = this.mobjCleanUpManager;

			/* 実験的に、カレントの移動も CreateChild() 内でやってみる。 */
			/* 駄目だったらまた直せばいいか？？？ */
			if ( enmMove == colonyCURRENT_MOVEMENT.TONEWER )
				this.mobjCleanUpManager.Current = objNew;

			return objNew;
		}



		internal clsDispDir CreateOrGetFunc( string sFuncName )
		{
			clsDispDir		objDispDir_Ret;
			objDispDir_Ret = this.marChildList.Find( x => ( x.FuncName == sFuncName) );
			if ( objDispDir_Ret != null ) {
				this.mobjCleanUpManager.Current = objDispDir_Ret;
				return objDispDir_Ret;
			}

			return this.CreateFunc( FuncName );
		}

		internal clsDispDir CreateFunc( 
						string sFuncName 
					)
		{

			clsDispDir			objNew;

			#if __DEBUG__
				objNew = this.GetChild_FromFuncName( sFuncName );
				if ( objNew != null ) {
					clscsPub.PErrMsg( "\"%s\" is already added.", sFuncName );
					return objNew;
				}
			#endif

			objNew = this.CreateChild();
			objNew.mbFunc						= true;
			objNew.mbClassDispose				= false;
			objNew.FuncName						= sFuncName;
			objNew.mobjCurrentClassInstance		= null;
			objNew.mtypClass					= null;

			return objNew;
		}

		internal clsDispDir CreateChildModule( 
						object objClassInstance, 
						string sModuleName, 
						System.Type typClassName
					)
		{
			clsDispDir			objNew;
			objNew = this.CreateChild( colonyCURRENT_MOVEMENT.NO_MOVE );
			objNew.mbFunc						= false;
			objNew.mbClassDispose				= true;
			objNew.FuncName						= sModuleName;
			objNew.mobjCurrentClassInstance		= objClassInstance;
			objNew.mtypClass					= typClassName;


			return objNew;
		}

		internal bool RemoveChild( clsDispDir objInstance )
		{
			bool bRet;
			bRet = this.marChildList.Remove( objInstance );
			clscsPub.Assert( bRet, "not found object to remove! program error. " );
			return bRet;
		}



		internal clsDispDir GetChild_FromClassInstance( object objInstance ) 
		{
			foreach( clsDispDir objIndex in this.marChildList )
			{
				if ( System.Object.ReferenceEquals( 
							objIndex.GetModuleInstance(),
							objInstance ) == true ) {

					return objIndex;
				}
			}
			return null;
		}

		internal clsDispDir GetChild_FromFuncName( string sFuncName ) 
		{
			// わけわかんねぇわｗ
			foreach( clsDispDir objIndex in this.marChildList )
			{
				if ( objIndex.FuncName == sFuncName ) {
					return objIndex;
				}
			}
			return null;
		}


		internal clsDispDir GetChild_FromClassTypedef( System.Type typClassDef ) 
		{
			// わけわかんねぇわｗ
			foreach( clsDispDir objIndex in this.marChildList )
			{
				if ( objIndex.ClassType == typClassDef ) {
					return objIndex;
				}
			}
			return null;
		}

		internal clsMultiDisposer MultiDisposer
		{
			set { this.mobjMDisposer = value; }
			get { return this.mobjMDisposer; }
		}


		internal System.Collections.Generic.List<stCleanup> CleanUpList
		{
			set { this.mobjMDisposer.CleanUpList = value; }
			get { return this.CleanUpList; }
		}


		internal bool IsShowDisposeExection
		{
			set {
				if ( this.mobjMDisposer == null )
					this.mobjMDisposer = new clsMDisposer_With_UpperMovement( this );

				this.mobjMDisposer.IsShowDisposeExection = value;
			}
			get {
				if ( this.mobjMDisposer == null )
					this.mobjMDisposer = new clsMDisposer_With_UpperMovement( this );

				return this.mobjMDisposer.IsShowDisposeExection;
			}

		}



		internal clsCleanUpDirManager Manager
		{
			set {
				this.mobjCleanUpManager = value;
			}
			get {
				return this.mobjCleanUpManager;
			}

		}





		// ファンクション実行中にエラー：エラーロールバック クリーンナップ
		//                       成功 ：モジュールレベルのalways メモリ開放クリーンナップ
		// が出来る、コストの低い関数を定義開発する。

	#endregion			//			clsDispDir:property / CreateChild   ( ParentModule/ParentFunc/GetParent etc  )





	#region clsDispDir: Add_CleanupWhenAlways() / Add_CleanupWhenOnlyErrRollback() / Add_CleanupForModuleVariable() etc

		// clsDispDir は「特定関数」用のスコープのみをフォローするclenaupリストの保存クラスとする。
		//	かつ、中身は System.IDisposable の配列形式のみ保持する(実際には generic List )。

		// clsMultiDisposer へのアクセスは、原則、clsDispDir を介しても追加/削除出来るようにする。
		
		// clsDispDir では１つの clsMultiDisposer インスタンスのみ情報を持てるようにする。
		// １つの cleanUp 登録で、複数の Dispose を一度に行いたい場合は clsDisposerArray で配列化して
		// そのインスタンスを cldDispDir に登録するようにする。
		//   ( clsDispDir メンバ mobjMDisposer を clsMultiDisposer から List<clsMultiDisposer> にしても
		//     良いのだが、
		//	   CleanUp_sucucessfully_Once で複数の dispose を一度にやりたい、という状況が
		//	   稀ではないかと )

		public void Add_CleanupWhenAlways( System.IDisposable objDispoer ) {
			/* 継承インターフェイス */
			this.Add_CleanupWhenAlways( objDispoer, null );
		}


		internal void Add_CleanupWhenAlways( System.IDisposable objDispoer, string sListName ) {

			// ・・・・.finaliize が実装されているかどうかは
			// 気にしないでほっとこう。

			// bool 値は、エラーロールバック時にのみ実行する時は true
			//				常にクリーンアップされる場合は     false

			if ( this.mobjMDisposer == null )
				this.mobjMDisposer = new clsMDisposer_With_UpperMovement( this );

			this.mobjMDisposer.Add_CleanupWhenAlways( objDispoer, sListName );
		}

		public void Add_CleanupWhenOnlyErrRollback( System.IDisposable objDispoer ) {
			/* 継承インターフェイス */
			this.Add_CleanupWhenOnlyErrRollback( objDispoer, null );
		}

		internal void Add_CleanupWhenOnlyErrRollback( System.IDisposable objDispoer, string sListName ) {


			if ( this.mobjMDisposer == null )
				this.mobjMDisposer = new clsMDisposer_With_UpperMovement( this );

			this.mobjMDisposer.Add_CleanupWhenOnlyErrRollback( objDispoer, sListName );
		}


		// 何に使ってんだろこれ・・・分からん・・・
		//		private void Add_stCleanup( System.IDisposable objDispoer, string sListName, bool bIsErrorCleanUp ) {
		//			stCleanup					stpCleanup;
		//			stCleanup.colonyWHENDO		enmWhen;
		//		
		//			// bool 値は、エラーロールバック時にのみ実行する時は true
		//			//				常にクリーンアップされる場合は     false
		//			stpCleanup = new stCleanup( stCleanup.colonyWHENDO.ALLWAYS,  objDispoer, sListName  );
		//		
		//			this.mobjMDisposer.Add_CleanupWhenAlways( stpCleanup );
		//		
		//		
		//		}



		internal void Add_CleanupForModuleVariable( 
						System.IDisposable objDispoer, 
						clsDispDir objDisposerDir_CurentExectingFunction_Root, 
						clsDispDir objLastDir, 
						stCleanup.dlgtDisposeCallBack pNullSet )
		{
			this.Add_CleanupWhenOnlyErrRollback_ToUpperModuleWhenSuccess( 
						objDispoer, 
						null,
						objDisposerDir_CurentExectingFunction_Root, 
						objLastDir, 
						pNullSet );
		}

		internal void Add_CleanupForModuleVariable( 
						System.IDisposable objDispoer, 
						string sListName,
						clsDispDir objDisposerDir_CurentExectingFunction_Root, 
						clsDispDir objLastDir, 
						stCleanup.dlgtDisposeCallBack pNullSet )
		{
			// 関数名が長いのでエリアスする
			this.Add_CleanupWhenOnlyErrRollback_ToUpperModuleWhenSuccess(
						objDispoer, 
						sListName,
						objDisposerDir_CurentExectingFunction_Root, 
						objLastDir,
						pNullSet );
		}


		internal void Add_CleanupWhenOnlyErrRollback_ToUpperModuleWhenSuccess( 
						System.IDisposable objDispoer, 
						string sListName,
						clsDispDir objDisposerDir_CurentExectingFunction_Root, 
						clsDispDir objLastDir, 
						stCleanup.dlgtDisposeCallBack pNullSet ) {
			// エラーロールバック時は通常通り、ErrorCleanUp で
			// モジュール変数の消去を行う。
			// 成功時は自身の親ディレクトリにて、
			// エラークリーンナップ時：消去
			// 通常クリーンナップ時  ：さらに上の階層に移動
			//							最上位ファンクションならモジュール変数 Dispose() に
			//							移動
			
			stCleanup			stpCleanup;

			stpCleanup = new stCleanup( stCleanup.colonyWHENDO.ONLYROLLBACK,  objDispoer, sListName  );
			stpCleanup.mbMoveToParent_WhenSuccess	= true;
			stpCleanup.mobjDir_WhenRootFuncSuceed	= objLastDir;
			stpCleanup.mobjDisposerDir_CurentExectingFunction_Root	=
						 objDisposerDir_CurentExectingFunction_Root;
			stpCleanup.pCallBack_NullSet = pNullSet;


			if ( this.mobjMDisposer == null )
				this.mobjMDisposer = new clsMDisposer_With_UpperMovement( this );

			this.mobjMDisposer.Add_stCleanup( stpCleanup );
		}

		public void Add_CleanupWhenOnlySuccessEnd( System.IDisposable objDispoer ) {
			/* 継承インターフェイス */
			this.Add_CleanupWhenOnlySuccessEnd( objDispoer, null );
		}

		internal void Add_CleanupWhenOnlySuccessEnd( System.IDisposable objDispoer, string sListName ) {
			stCleanup			stpCleanup;

			stpCleanup = new stCleanup( stCleanup.colonyWHENDO.ONLYSUCCUESSEND,  objDispoer, sListName  );


			if ( this.MultiDisposer == null )
				this.MultiDisposer = new clsMDisposer_With_UpperMovement( this );

			this.MultiDisposer.Add_stCleanup( stpCleanup );
		}

		internal void Add_CleanupWhen( System.IDisposable objDisposable, stCleanup.colonyWHENDO nWhenDo ) {
			this.Add_CleanupWhen( objDisposable, null, nWhenDo );
		}

		internal void Add_CleanupWhen( System.IDisposable objDisposable, string sListName, stCleanup.colonyWHENDO nWhenDo ) {

			if ( nWhenDo == stCleanup.colonyWHENDO.ONLYROLLBACK) {
				this.Add_CleanupWhenOnlyErrRollback( objDisposable, sListName );
			} else if ( nWhenDo == stCleanup.colonyWHENDO.ONLYSUCCUESSEND ) {
				this.Add_CleanupWhenOnlySuccessEnd( objDisposable, sListName );
			} else {
				this.Add_CleanupWhenAlways( objDisposable, sListName );

			}

		}		//		void Add_CleanupWhen()





		internal void Add_CleanupForModuleVarEx( 
						System.IDisposable objDispoer, 
						string sListName,
						clsDispDir objDisposerDir_CurentExectingFunction_Root, 
						clsDispDir objLastDir, 
						stCleanup.dlgtDisposeCallBack pNullSet )
		{
			// 削除機能つきのクリーンナップ追加
			
			bool			bDeleted = false;
			clsDispDir		objDispDir_Target = this;
			while ( true ) {

				int nIndex;
				nIndex = objDispDir_Target.MultiDisposer.
						GetDisposerIndexFromListName( sListName );
				if ( nIndex >= 0 ) {
					/* なぜにエラークリーンナップ・・・ */
					/* 正常クリーンナップでも良さそうだけど、よく分からんわ。 */
					objDispDir_Target.MultiDisposer.ErrorCleanUp( nIndex );
					bDeleted = true;
					break;
				}


				objDispDir_Target = objDispDir_Target.Parent;

				if ( objDispDir_Target == null )
					break;
				
				clscsPub.Assert( objDispDir_Target != null, "objDispDir_Target is null!" );

			}

			if ( !bDeleted ) {
				int nIndex;
				nIndex = objLastDir.MultiDisposer.
						GetDisposerIndexFromListName( sListName );

				if ( nIndex >= 0 ) {
					/* なぜにエラークリーンナップ・・・ */
					/* 正常クリーンナップでも良さそうだけど、よく分からんわ。 */
					objLastDir.MultiDisposer.ErrorCleanUp( nIndex );
				}
			}

			if ( this.mobjMDisposer == null )
				this.mobjMDisposer = new clsMDisposer_With_UpperMovement( this );

			this.Add_CleanupWhenOnlyErrRollback_ToUpperModuleWhenSuccess(
						objDispoer, 
						sListName,
						objDisposerDir_CurentExectingFunction_Root, 
						objLastDir,
						pNullSet );

		}		//		void Add_CleanupForModuleVarEx()




	#endregion			//		clsDispDir: Add_CleanupWhenAlways() / Add_CleanupWhenOnlyErrRollback() / Add_CleanupForModuleVariable() etc



	#region clsDispDir: ErrorCleanUp() / SuccessEndCleanUp() / CleanUpChildren() 
		public void ErrorCleanUp()
		{
			/* インターフェイス継承 */
			this.CleanUpChildren( this, true );
		}

		public void SuccessEndCleanUp()
		{
			/* インターフェイス継承 */
			this.CleanUpChildren( this, false );
		}

		private void CleanUpChildren( clsDispDir objDirData, bool bErrorCleanUp )
		{
			if ( objDirData == null )
				return;



			if ( objDirData.HasChild() ) {
				// 子供から全部クリーンナップ
				System.Collections.Generic.List< clsDispDir >	lstChildren;

				lstChildren = objDirData.GetChildList();
				for( int nIndex = lstChildren.Count - 1; nIndex >= 0; nIndex-- ) {
					clsDispDir		 objIndex = 
						lstChildren[ nIndex ];

					if ( objIndex == null )
						continue;

					this.CleanUpChildren( objIndex, bErrorCleanUp );
				}
			}


			if ( this.mobjMDisposer != null ) {

				if ( bErrorCleanUp )
					objDirData.mobjMDisposer.ErrorCleanUp();
				else
					objDirData.mobjMDisposer.SuccessEndCleanUp();


				// 空だとは思うけど念のため呼んでおく
				objDirData.mobjMDisposer.Dispose();
				objDirData.mobjMDisposer = null;
			}


			// 親から current を抜く
			clsDispDir		objParent;
			objParent = objDirData.Parent;
			if ( objParent != null ) {
				objParent.RemoveChild( objDirData );
			}

			this.MoveUpCurrentDir();		//	 マネージャのカレント位置を自動で変更。いるかこれ？

			objDirData.Parent = null;

		}		//		void CleanUpChildren()


		private void  MoveUpCurrentDir() {
			if ( this.mobjCleanUpManager == null )		return;

			if ( this.mobjCleanUpManager.Current == this ) {
				/* カレントの位置を自動的に変更しておく。 */
				/* いるかこの処理？？？怪しくならないか？ */
				this.mobjCleanUpManager.Current = this.mobjParent;
			}
		}		//	void  MoveUpCurrentDir()


		public void CleanUp_sucucessfully_Once()
		{
			/* インターフェイス継承 */
			if ( this.mobjMDisposer == null )
				return;

			this.mobjMDisposer.CleanUp_sucucessfully_Once();


			if ( this.mobjMDisposer.IsNoCleanUpList() ) {
				this.mobjMDisposer.Dispose();
				this.mobjMDisposer = null;
			}

		}





	#endregion		//		clsDispDir: ErrorCleanUp() / SuccessEndCleanUp() / CleanUpChildren() 





}		//		class clsDispDir












internal class clsMDisposer_With_UpperMovement : clsMultiDisposer {

	/* ！！ 「無用に変数参照をすると、ガベージコレクタにより破棄されにくくなる」 */
	/*     という概念を追加し、使わなくなったポインタは極力すぐ null セット */
	/*     しまくる、という概念を追加してみる。 */

	/* C# では、new / override による関数継承で動きがかなり変わるらしい。 */
	/* 継承クラスインスタンスを基底クラスでポインタを受けた時、new は基底クラスを、virtual/overwrite した関数では */
	/* を呼び出すらしい。なんじゃそりゃ。 */

	/* MultiDisposer 上に、上位側移動機能を入れるのは何かおかしい気がする。 */
	/* ので、一応分離しておく。 */

	private clsDispDir			mobjDispDir_Cuurent;

	internal clsMDisposer_With_UpperMovement() : base()
	{
		this.mobjDispDir_Cuurent = null;
	}

	internal clsMDisposer_With_UpperMovement( clsDispDir objDispDir_Current ) : base()
	{
		this.mobjDispDir_Cuurent = objDispDir_Current;
	}

	internal clsDispDir			CurrentDisposerDir
	{
		set {		this.mobjDispDir_Cuurent = value; }
		get {		return this.mobjDispDir_Cuurent; }
	}

	protected internal override void MoveUpTo_DispDirOrModule( ref stCleanup stpMove )
	{
		bool bIsThis;

		bIsThis = System.Object.ReferenceEquals( this, 
						stpMove.mobjDisposerDir_CurentExectingFunction_Root );

		if ( bIsThis ) {
			// モジュール Dispose へ移動
			stpMove.mobjDir_WhenRootFuncSuceed.
				Add_CleanupWhenAlways( stpMove.mobjDisposable );
		} else {
			// 親のエラークリーンナップに配置変更


			// ここは何書いてあるか分からんから、直接構造体をセットする方法に変えた方がいいかも。

			clscsPub.Assert( this.mobjDispDir_Cuurent.Parent != null, 
								"not unset parent pointer . " );
			this.mobjDispDir_Cuurent.Parent.
				Add_CleanupForModuleVariable(
							stpMove.mobjDisposable, 
							stpMove.mobjDisposerDir_CurentExectingFunction_Root, 
							stpMove.mobjDir_WhenRootFuncSuceed,
							stpMove.pCallBack_NullSet
						);
		}


		return;
	}		//		void MoveUpDispDir_Or_MoeveToModule()



}	//		class clsMDisposer_With_UpperMovement



internal class clsMultiDisposer : Interface_MultiDisposer {


	// 以下、旧 MultiDisposer メンバ変数
	// dispose が行われた時に、開放された処理やクラス名等を表示するかどうか。
	private bool											mbShowExection;

	private bool											mbDisposeIsErrCleanUp;
	private System.Collections.Generic.List<stCleanup>		marStackList;



	#region clsMultiDisposer: constructor / Dispose()
	

		internal clsMultiDisposer() {
			this.marStackList = null;
			this.mbShowExection = false;
			this.mbDisposeIsErrCleanUp = false;

		}
		
		internal clsMultiDisposer( bool bDisposeIsErrCleanUp ) {
			this.marStackList = null;
			this.mbShowExection = false;

			// dispose() が呼ばれた時にエラークリーンナップを呼び出すなら true,
			// 正常クリーンナップを呼び出すなら false
			this.mbDisposeIsErrCleanUp = bDisposeIsErrCleanUp;
		}


		public void Dispose() {
			// 消滅時、クリーンナップが残ってたら実行
			if ( this.marStackList == null )	return;

			if ( this.marStackList.Count != 0 ) {
			
				if ( this.mbDisposeIsErrCleanUp)
					this.ErrorCleanUp();
				else
					this.SuccessEndCleanUp();

			}

			this.marStackList.Clear();
			this.marStackList = null;
		}

	#endregion		//		clsMultiDisposer: constructor / Dispose()



	#region clsMultiDisposer: property
		internal System.Collections.Generic.List<stCleanup> CleanUpList
		{
			set { this.marStackList = value; }
			get { return this.marStackList; }
		}

		internal bool IsNoCleanUpList()
		{
			if ( this.marStackList == null)
				return true;

			return ( this.marStackList.Count == 0);
		}


		internal bool IsShowDisposeExection
		{
			set {
				this.mbShowExection = value;
			}
			get {
				return this.mbShowExection;
			}

		}
	#endregion




	#region clsMultiDisposer: Add_CleanupWhenAlways() / Add_CleanupWhenOnlyErrRollback() / Add_CleanupForModuleVariable() etc



		public void Add_CleanupWhenAlways( System.IDisposable objDispoer ) {
			/* インターフェイス継承 */
			this.Add_CleanupWhenAlways( objDispoer, null );
		}


		internal void Add_CleanupWhenAlways( System.IDisposable objDispoer, string sListName ) {
			this.Add_Disposable( objDispoer, sListName, false );
		}

		public void Add_CleanupWhenOnlyErrRollback( System.IDisposable objDispoer ) {
			/* インターフェイス継承 */
			this.Add_CleanupWhenOnlyErrRollback( objDispoer, null );
		}

		internal void Add_CleanupWhenOnlyErrRollback( System.IDisposable objDispoer, string sListName ) {
			this.Add_Disposable( objDispoer, sListName, true );
		}

		private void Add_Disposable( System.IDisposable objDispoer, string sListName, bool bIsExecutedWhenOnlyErrorCleanUp )
		{
			stCleanup					stpCleanup;
			stCleanup.colonyWHENDO		enmWhen;

			if ( bIsExecutedWhenOnlyErrorCleanUp )
				enmWhen = stCleanup.colonyWHENDO.ONLYROLLBACK;
			else
				enmWhen = stCleanup.colonyWHENDO.ALLWAYS;

			stpCleanup = new stCleanup( enmWhen,  objDispoer, sListName  );


			if (this.marStackList == null )
				this.marStackList = new System.Collections.Generic.List<stCleanup>();
			
			this.marStackList.Add( stpCleanup );

		}

		public void Add_CleanupWhenOnlySuccessEnd( System.IDisposable objDispoer ) {
			/* インターフェイス継承 */
			stCleanup					stpCleanup;

			stpCleanup = new stCleanup( stCleanup.colonyWHENDO.ONLYSUCCUESSEND, 
								 objDispoer, null  );


			if (this.marStackList == null )
				this.marStackList = new System.Collections.Generic.List<stCleanup>();
			
			this.marStackList.Add( stpCleanup );
		}


		internal void Add_stCleanup( stCleanup stpCleanUp ) {

			if (this.marStackList == null )
				this.marStackList = new System.Collections.Generic.List<stCleanup>();
			
			this.marStackList.Add( stpCleanUp );

		}


	#endregion			//		clsDispDir: Add_CleanupWhenAlways() / Add_CleanupWhenOnlyErrRollback() / Add_CleanupForModuleVariable() etc



	#region clsMultiDisposer: ErrorCleanUp() / ErrorCleanUp_Once() / SuccessEndCleanUp() / CleanUp_sucucessfully_Once()

		public void ErrorCleanUp() {
			/* インターフェイス継承 */
			while ( this.marStackList.Count > 0 ) {
				this.ErrorCleanUp_Once();
			}
		}

		internal void ErrorCleanUp( int nIndex ) {
			this.ErrorCleanUp_Once_withIndex( nIndex );
		}


		private void ErrorCleanUp_Once() {

			if ( this.marStackList == null )
				return;
			if ( this.marStackList.Count == 0 )
				return;


			this.ErrorCleanUp_Once_withIndex( this.marStackList.Count - 1 );
		}


		private void ErrorCleanUp_Once_withIndex( int nIndexc ) {

			if ( this.marStackList.Count == 0 )
				return;

			stCleanup				stpCleanup;
			System.IDisposable		objDisposable;


			stpCleanup    = this.marStackList[ nIndexc ];
			// 常に実行される、でいいんじゃねーの？
			if ( stpCleanup.IsExecOnlyError() || 
				stpCleanup.IsExecAllways() ) {

				objDisposable = stpCleanup.mobjDisposable;

				this.ShowDisposableName( ref stpCleanup );

				objDisposable.Dispose();


				/* オブジェクト変数 null セットも一応入れとくか。 */
				if ( stpCleanup.pCallBack_NullSet != null ) {
					stpCleanup.pCallBack_NullSet();
				}

			}
			this.marStackList.RemoveAt( nIndexc );


			// stpCleanUp を破棄
			stpCleanup.mobjDisposable              = null;
			stpCleanup.mobjDir_WhenRootFuncSuceed  = null;
			stpCleanup.mobjDisposerDir_CurentExectingFunction_Root = null;


		}

		public void SuccessEndCleanUp() {
			/* インターフェイス継承 */
			while ( this.marStackList.Count > 0 ) {
				this.CleanUp_sucucessfully_Once();
			}
		}


		// エラーロールバックで、スタック一個だけ
		//	Dispose する、ってパターンは今んとこ無いらしいね。
		//internal void CleanUp_Once() {
		//}

		public void CleanUp_sucucessfully_Once() {
			/* インターフェイス継承 */

			// 正常終了処理のメモリクリンナップチェーンのうち
			// スタック積みされた一番「上」の１つだけを実行


			/* '24.05.31 追記：げ・・・・親情報がここで取れない。 */
			/*  MoveToParent 動作時に、親の clsDispDir が分からないので移動出来ない。 */
			/* clsMultiDisposer 単独から cleanup 処理を呼ばれた際には、 */
			/* 「親に移動する」という処理は出来なくした方が良さそうかも？？？ */
			/* つまり、「親に MultiDisposer を付け替える」という処理は、 */
			/* MultiDisposer としての機能から一歩はみ出ているので、 */
			/* 別の継承クラスで出来るようにオーバーライドするか、 */
			/* clsDispDir 専用機能とするのが妥当では。 */

			/* MultiDisposer 単独では、あくまでも、単独で複数登録された Dispose() */
			/* 処理を呼び出すのみ、とするのが妥当だろう。 */


			if ( this.marStackList.Count == 0) {
				return;
			}
			stCleanup				stpCleanup;
			System.IDisposable		objDisposable;


			stpCleanup    = this.marStackList[ this.marStackList.Count - 1 ];
			bool bExec = false;
			if ( stpCleanup.IsExecOnlyError() == true ) {
				bExec = false;
			} else if ( stpCleanup.IsExecOnlySucessEnd() == true ) {
				bExec = true;
			} else {
				// always 
				bExec = true;
			}

			if ( stpCleanup.mbMoveToParent_WhenSuccess ) {
				bExec = false;
			}

			if ( bExec == true ) {
				objDisposable = stpCleanup.mobjDisposable;

				this.ShowDisposableName( ref stpCleanup );

				objDisposable.Dispose();

				if ( stpCleanup.pCallBack_NullSet != null ) {
					// デレゲート呼び出しの際に、Invoke 経由で呼ばなくて良いの？？
					//
					// 別スレッドからUI描画処理を含めて依頼する訳でもないので、
					// OnPaint 処理を行う訳でもないので、
					// invoke を使わんでもいいらしい。普通にデレゲート呼び出しで
					// いいらしい。
					stpCleanup.pCallBack_NullSet();
				}
				stpCleanup.pCallBack_NullSet = null;

			} else {
				// クリーンナップを移動する処理が必要なのでは？ここは？

			}



			this.marStackList.RemoveAt( this.marStackList.Count - 1 );

			// 成功時、このクリーンナップを上位側
			// エラー時クリーンナップシーケンスに追加する必要があるなら
			// 上位側に移転する。
			// 現階層が最上位で成功クリーンナップ中なら
			// モジュール変数に付け替える。。
			if ( stpCleanup.mbMoveToParent_WhenSuccess ) {
				this.MoveUpTo_DispDirOrModule( ref stpCleanup );
			}


			// stpCleanUp を破棄
			stpCleanup.mobjDisposable              = null;
			stpCleanup.mobjDir_WhenRootFuncSuceed  = null;
			stpCleanup.mobjDisposerDir_CurentExectingFunction_Root = null;

		}

		protected internal virtual void MoveUpTo_DispDirOrModule( ref stCleanup stpMove )
		{
			/* ref 構造体渡しにすれば、ポインタ渡しと同じ動作になるかしらん？？ */
			return;
		}


	#endregion			//		clsMultiDisposer: HasCleanUp_FromListName() など


	internal bool HasCleanUp_FromListName( string sListName ) {
		foreach ( stCleanup stpCurrent in marStackList )
		{
			if ( stpCurrent.msName == sListName ) {
				return true;
			}

		}	//	foreach
		return false;
	}	//	bool List_HasCleanUp_FromListName()




	private void ShowDisposableName( ref stCleanup stpCurrent ) {

		System.IDisposable		objDisposable;

		objDisposable = stpCurrent.mobjDisposable;
		if ( !this.mbShowExection ) 
			return;

		if ( objDisposable is clsCallBackDisposer ) {
			clsCallBackDisposer		objCbDisposer 
				= (clsCallBackDisposer)objDisposable;
			string sName;
			if ( objCbDisposer != null )
				sName = objCbDisposer.FuncName;
			else
				sName = objCbDisposer.GetType().FullName;

			clscsPub.PErrMsg( "callback diposer is called now: name {0}. ", 
					sName );
			
			return;
		}

		System.Type		typType;
		if ( objDisposable is clsComDisposer ) {
			clsComDisposer		objComDisposer
				= (clsComDisposer)objDisposable;

			typType = objComDisposer.csType;
			if ( typType == null) {
				object		objFromDisposable;
				objFromDisposable = ((clsComDisposer)objDisposable).WrappedCom;
				typType = objFromDisposable.GetType();
			}

		} else if ( stpCurrent.msName != null ) {
			
			clscsPub.PErrMsg( "dipose is called now: ListName \"{0}\". ", 
							stpCurrent.msName );
			return;

		} else {
			object			objFromDisposable;
			objFromDisposable = objDisposable;
			typType = objFromDisposable.GetType();
		}


		clscsPub.PErrMsg( "dipose is called now: type {0}. ", 
				typType.FullName );
	}







	#region 特殊なクリーンナップの登録削除用の関数
		// 特殊なクリーンナップ削除用関数
		//		何に使ってたっけ・・・。

		internal int  GetDisposersIndex( System.IDisposable objDisposable )
		{
			if ( this.marStackList.Count == 0 )
				return -1;


			
			for ( int nIndex = this.marStackList.Count - 1; nIndex >= 0; nIndex-- ) {
				stCleanup			stpCurrent;

				stpCurrent = this.marStackList[ nIndex ];
				if ( System.Object.ReferenceEquals( stpCurrent.mobjDisposable, objDisposable ) == true )
				{
					return nIndex;
				}
			}
			return -1;
		}

		internal void DeleteDisposer( int nIndex )
		{
			this.marStackList.RemoveAt( nIndex );
		}

		internal int GetDisposerIndexFromListName( string sListName )
		{
			if ( this.marStackList.Count == 0 )
				return -1;
			

			int nFind;

			/* デレゲート */
			/* 匿名メソッド */
			/* 書き方よく分からん・・・・ */
			//stpCurrent = this.marStackList.
			//		Find(delegate (stCleanup n) { return stpCurrent.msName = sListName; });


			/* ラムダ式 */
			nFind = this.marStackList.
					FindIndex( stpCurrent => stpCurrent.msName == sListName );

			/* 上記を以下のようにしていたため、15分ぐらい?、ちょっとハマってしまった・・・ */
			/*		FindIndex( stpCurrent => stpCurrent.msName = sListName ); */

			return nFind;
		}

	#endregion			//		特殊なクリーンナップの登録削除用の関数


}		//		class clsMultiDisposer



	/**********************************************/
	/*   Disposer 補助クラス                      */
	/**********************************************/
	#region 	MultiDisposer 関連以外の補助 Dispose クラス

		internal class clsComDisposer : System.IDisposable
		{
			private object					mobjCom;
			private System.Type				mtypInterface;

			internal object WrappedCom {
				set {	this.mobjCom = value; }
				get {	return this.mobjCom; }
			}


			internal System.Type csType {
				set {	this.mtypInterface = value; }
				get {	return this.mtypInterface; }
			}


			internal clsComDisposer( object objCom )
			{
				this.mobjCom = objCom;
				this.mtypInterface = null;
			}

			internal clsComDisposer( object objCom, System.Type typInterface )
			{
				this.mobjCom = objCom;
				this.mtypInterface = typInterface;
			}


			public void Dispose() {
				/* インターフェイス継承 */
				if ( this.mobjCom != null ) {
					System.Runtime.InteropServices.Marshal.ReleaseComObject( this.mobjCom );
					this.mobjCom = null;
				}
			}


		}		//		class clsComDisposer



		internal class clsCallBackDisposer : System.IDisposable
		{
			internal delegate void dlgtDisposeCallBack();

			internal string							msName;
			private dlgtDisposeCallBack				mpdlgtDisposeCallBack;

			internal string FuncName {
				set { this.msName = value; }
				get { return this.msName; }
			}

			internal clsCallBackDisposer( string sName, dlgtDisposeCallBack pFunc )
			{
				this.msName = sName;
				this.mpdlgtDisposeCallBack = pFunc;
			}



			public void Dispose() {
				/* インターフェイス継承 */
				if ( this.mpdlgtDisposeCallBack != null ) {
					this.mpdlgtDisposeCallBack();
				}
			}


		}		//		class clsComDisposer


		internal class clsCleanUpMovement_Disposer : System.IDisposable
		{
			// クリーンナップ処理において、対象となるクリーンナップ処理自体を
			// 異なるクリーンナップチェーンへと所属変更させる

			// 主に、 clsCLRManager :: mobjClrv4Debugging を、
			//		clsCsDebugger :: AttachExe() 処理中に
			//		 新規作成したタスクが正常成功した場合に
			//		mobjClrv4Debugging はモジュール clsCLRManager:: mobjClrManager 
			//		が開放されるまで、または DettachExe() が呼び出されるまで
			//		保持される。clsCLRManager:: mobjClrManager が開放== Dispose()
			//		処理を呼び出される時、そのメンバ変数 mobjClrv4Debugging は
			//		メモリ開放処理がされなくてはならない( COM オブジェクトなので)
			//		ので、その Marshal.ReleaseComObject ( mobjClrv4Debugging ) 
			// 		呼び出しを clsCLRManager::Disose() に
			//		追加しなくてはならない。
			//
			//		重複登録を避け、AttachExe から clsCLRManager::Disose() へと
			//		Marshal.ReleaseComObject ( mobjClrv4Debugging ) 　クリーンナップ
			//		処理を移動するのが当クラス clsCleanUpMovement_Disposer の
			//		目的。
			//		それ以外の用途でも使う事になるかどうかは今後の様子見という
			//		事で・・・。
			//		

			//		clsDispDir の SuccessEndCleanUp 
			//		自体に引っ越し処理を入れたので、このクラスは
			//		とりあえず使われていない。

			private clsDispDir		monjCurrentParent, monjNewParent;
			private System.IDisposable		mobjDisposable;

			internal clsDispDir CurrentParent {
				set { this.monjCurrentParent = value; }
				get { return this.monjCurrentParent; }
			}

			internal clsDispDir NewParent {
				set { this.monjNewParent = value; }
				get { return this.monjNewParent; }
			}


			internal System.IDisposable Disposer {
				set { this.mobjDisposable = value; }
				get { return this.mobjDisposable; }
			}


			//	クラス名候補：
			// transfer
			// take out
			// plant
			// move
			// export
			// reset
			// reattach
			// Adopt a child   養子に出す
			// change address  住所変更？
			// 

			public void Dispose()
			{
				/* インターフェイス継承 */

				// 正常終了時：ここが実行される。
				// エラー実行時はこのチェーンは無視される。

				//  --->> ?? あれ？メモリリークしてんじゃね？
				//		エラー実行時は、「この引っ越しタスクを削除する」という
				//		処理が必要になんじゃね？？？
				//		実行無視されるだけだし、別に問題はないはずじゃ？？
				//		だいたい、親ディスポーザーにここからアクセス出来ないし・・・。

				clsMultiDisposer	objTarget;
				objTarget = monjCurrentParent.MultiDisposer;

				// Disposer 引っ越し
				int		nCneanUpIndex;
				nCneanUpIndex = objTarget.GetDisposersIndex(
								this.mobjDisposable );

				clscsPub.Assert( nCneanUpIndex >= 0, "not found match Disposable . " );

				stCleanup		stpClenup;

				/* 登録されている stCleanup 構造体をコピー */
				stpClenup = objTarget.CleanUpList[ nCneanUpIndex ];

				objTarget.DeleteDisposer( nCneanUpIndex );



				this.monjNewParent.MultiDisposer.Add_stCleanup( stpClenup );

			}		// void Dispose()

		}	//		class clsCleanUpMovement_Disposer



		internal class clsDisposerArray : System.IDisposable {
			// MultiDisposer    Dispose() の配列
			//	１つの cleanup タスクに、複数の dispose 処理をパッキングするための
			//	追加クラス
			//
			//		これ無いと clsDispDir の保持リストそさらに多重配列化する必要が
			//		出て来る
			//

			internal System.Collections.Generic.List<System.IDisposable>		marDisposeArray;

			internal void Add( System.IDisposable	objDiposer )
			{
				if ( this.marDisposeArray == null ) {
					this.marDisposeArray = new System.Collections.Generic.List<System.IDisposable>();
				}
				this.marDisposeArray.Add( objDiposer );
			}

			public void Dispose()
			{
				/* インターフェイス継承 */

				foreach( System.IDisposable objDiposer in this.marDisposeArray )
				{
					if ( objDiposer != null ) {

						objDiposer.Dispose();
					}
				}	//	 foreach

			}	//		internal Dispose()

		}		//		class clsDisposerArray

	#endregion			//		MultiDisposer 関連以外の補助 Dispose クラス





}		// namespace CliDbger

